#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from scipy.special import erfinv,erf

# Own modules
from models import CBL
from observations import *

# TODO: comprehensive documentation is missing.

# TODO: code should already permit to perturb and/or estimate multiple parameters
#  But it needs testing.

# TODO: code should already permit assimilating u and v observations.
#  But it needs testing.

# TODO: add other options for variance inflation (multiplicative, additive).

# TODO: add other options for parameter transformations.

verbose = True

def transform_none(x,kind):
    if kind=='dir':
        return x
    if kind=='inv':
        return x

def transform_log(x,kind):
    if kind=='dir':
        return np.log(x)
    if kind=='inv':
        return np.exp(x)

def transform_pit(x,kind):
    # mu and sigma could be guessed from the data, but...
    a = 0.5
    b = 4.5
    mu = 0.5*(a+b)
    sigma = (b-a)/12**0.5
    # takes a uniform, returns a normal
    if kind=='dir': 
        p = (x-a)/(b-a)
        xt = mu+sigma*2**0.5*erfinv(2*p-1)
        return xt
    # takes a normal), returns a uniform
    if kind=='inv':
        p = 0.5*(1+erf((x-mu)/sigma/2**0.5))
        xt = a+p*(b-a)
        return xt

def observation_operator(fp,xp,x):
    f = np.interp(x, xp, fp)
    return f

def gaspari_cohn(r,cutoff):
    """Gaspari-Cohn function.
    https://github.com/ecosme38/Data-Assimilation-Notebooks/blob/master/Illustrate-Gaspari-Cohn-function-and-covariance-localization.ipynb
    """
    r = r.astype('float')
    ra = np.abs(r/cutoff)
    gp = np.zeros_like(r)
    
    i = (ra<=1.)
    gp[i] = -0.25*ra[i]**5+0.5*ra[i]**4+0.625*ra[i]**3-5./3.*ra[i]**2+1.
    
    i = (ra>1)&(ra<=2.)
    gp[i] = 1./12.*ra[i]**5-0.5*ra[i]**4+0.625*ra[i]**3+5./3.*ra[i]**2-5.*ra[i]+4.-2./3./ra[i]
    
    return gp

def compute_distances(xcoords,ocoords):
    # Number of state vector elements (coordinates are given)
    nstate = xcoords.size

    # Number of observations (coordinates are given)
    nobs = ocoords.size

    # Preallocate matrix to store distances between
    # pairs of state vector elements and observations
    dist = np.zeros((nstate,nobs)) + np.nan
    
    # i loops over state vector elements
    # j loops over observations
    for i in range(nstate):
        for j in range(nobs):
            dist[i,j] = np.abs(xcoords[i]-ocoords[j])
    
    # Return distances
    return dist

def eakf(xb,o,var_o,xcoord,ocoord,dist,\
                    localization_cutoff=np.infty,return_covariances_increments_and_innovations=False):

    def increments_in_obs_space(y_p,o,var_o):
    
        # Compute ensemble mean and ensemble variance of prior
        mean_p = np.mean(y_p)
        var_p = np.var(y_p)
    
        # Observation and prior weights
        weight_o = 1./var_o
        weight_p = 1./var_p
    
        # Compute weight of posterior, then ensemble variance of posterior
        weight_u =  weight_o + weight_p
        var_u = 1./weight_u

        # Shift the mean from mean_p to mean_u
        mean_u = (mean_p*weight_p + o*weight_o)*var_u 
        
        # Compute ensemble of analysis increments in observation space.
        # There are three terms in delta_y; the first two are the posterior
        # ensemble state (shifted mean + adjusted perturbations; e.g. Eq 19 in Anderson 2003);
        # the last one (y_p) is the prior state.
        # Their difference is the increment in observation space.
        alpha = (var_u*weight_p)**0.5
        #delta_y = mean_u*np.ones(nens) + (alpha-1)*y_p - alpha*mean_p*np.ones(nens) 
        delta_y = mean_u*np.ones(nens) + alpha*(y_p - mean_p*np.ones(nens)) - y_p
        
        return delta_y,weight_p
    
    # Read out relevant dimensions
    nvar = xb.shape[0]
    nens = xb.shape[1]
    nobs = o.shape[0]

    # Make a copy of the input array. Necessary in order to store correctly
    # the last forecast and the analysis.
    xu = np.copy(xb)

    # Compute unique values of distances and localization function.
    # This spares some computations because distances are discrete,
    # so a given value occurs more than once.
    ds = np.sort(np.unique(dist))
    loc = gaspari_cohn(ds,cutoff=localization_cutoff)

    # Preallocate array for model equivalents
    yp = np.zeros((nobs,nens))+np.nan
    increments = np.zeros((nvar,nobs))+np.nan

    # Preallocate array to save variances and covariances
    # (they are adjusted after ingesting every single observation).
    # This is not needed by the algorithm, but may be useful
    # for analysis purposes.
    if return_covariances_increments_and_innovations:
        cov_xx = np.zeros((nobs,nvar))+np.nan
        cov_xy = np.zeros((nobs,nvar))+np.nan
        cov_yy = np.zeros((nobs))+np.nan
        innovations_mean = np.zeros((nobs))+np.nan
        increments_mean = np.zeros((nobs,nvar))+np.nan
    
    # Process observations sequentially
    for i in range(nobs):

        # Apply observation operator to prior ensemble, i.e.
        # compute model equivalents of observations
        for j in range(nens):
            yp[i,j] = observation_operator(xu[:,j],xcoord,ocoord[i])

        # Store innovation wrt to ensemble mean
        if return_covariances_increments_and_innovations:
            innovations_mean[i] = o[i]-yp[i].mean()

        # Determine analysis increments in observation space
        # weight_p is a scalar
        # delta_y has dimension nens
        delta_y,weight_p = increments_in_obs_space(yp[i,:],o[i],var_o[i])

        # Update state variables in each ensemble member using ensemble covariances
        # The default in np.cov is N-1 degrees of freedom.
        for k in range(nvar):
            cov = np.cov(xu[k,:],yp[i,:]) # scalar
            # Compute analysis increments (=adjustments of ensemble perturbations)
            increments = cov[0,1]*weight_p*delta_y*\
                np.interp(dist[k,i],ds,loc) # nens values; second row is covariance localization
            xu[k,:] = xu[k,:] + increments
            if return_covariances_increments_and_innovations:
                cov_xx[i,k] = cov[0,0]
                cov_xy[i,k] = cov[0,1]
                cov_yy[i] = cov[1,1]
                increments_mean[i,k] = increments.mean()


    # Return
    if return_covariances_increments_and_innovations:
        return xu,yp,cov_xx,cov_xy,cov_yy,innovations_mean,increments_mean
    else:
        return xu,yp

def letkf(xb,o_global,var_o,xcoord,ocoord,dist,\
                     localization_cutoff=np.infty,inflation_rho=1.00):
    # Default size of local domain is set to 6 (=13 points), following Baek et al.    
    # Following strictly http://dx.doi.org/10.1016/j.physd.2006.11.008
    assert inflation_rho >= 1, \
        f"LETKF: Inflation factor must be > 1, got: {inflation_rho}"

    # Read out relevant dimensions
    nvar = xb.shape[0]
    nens = xb.shape[1]
    nobs = o_global.shape[0]

    # Create an empty array for the analysis.
    # Necessary in order to store correctly
    # the last forecast and the analysis.
    xa = np.zeros(xb.shape)+np.nan

    # Apply observation operator to prior ensemble, i.e.
    # compute model equivalents of observations
    yb_global = np.zeros((nobs,nens))+np.nan
    for i in range(nobs):
        for j in range(nens):
            yb_global[i,j] = observation_operator(xb[:,j],xcoord,ocoord[i])

    # Compute ensemble mean in observation space
    ybm_global = np.mean(yb_global,axis=1)

    # Compute matrix of ensemble perturbations in observation space
    Yb_global = yb_global-np.tile(ybm_global,(nens,1)).T

    # Compute ensemble mean in state space
    xbm = np.mean(xb,axis=1)

    # Compute matrix of ensemble perturbations in state space
    Xb = xb-np.tile(xbm,(nens,1)).T
    
    # Loop over grid points
    for m in range(nvar):

        # Get local model equivalents
        ybm = ybm_global[dist[m,:]<=localization_cutoff]
        Yb = Yb_global[dist[m,:]<=localization_cutoff]

        # Get local observations and their error variances
        yo = o_global[dist[m,:]<=localization_cutoff]
        vo = var_o[dist[m,:]<=localization_cutoff]

        # Form R matrix
        R = np.diag(vo)

        # Compute C matrix
        # The following statements are equivalent, the first is slightly faster
        # C = Yb.T@np.linalg.inv(R)
        # C = np.linalg.solve(R,Yb).T
        C = Yb.T@np.linalg.inv(R)

        # Compute analysis error variance matrix Pa
        Pa = np.linalg.inv( (nens-1)*np.eye(nens)/inflation_rho + C@Yb )
        
        # Compute weight matrix (symmetric square root of Pa)
        # Can use eigh and evectors.T because Pa is real symmetric
        evalues, evectors = np.linalg.eigh((nens-1)*Pa)
        Wa = evectors * np.sqrt(evalues) @ evectors.T
        
        # Compute ensemble mean weights
        wam = Pa @ C @ (yo-ybm)

        # Compute weights of ensemble perturbations
        Wai = Wa + np.tile(wam,(nens,1)).T
        
        # Compute local analysis ensemble members and store them
        # The ensemble mean would be
        # xam[m,:] = xbm[m] + Xb[m,:]@wam
        # but it's not necessary
        xa[m,:] = np.tile(xbm[m],(nens,1)).T + Xb[m,:] @ Wai
        
    return xa,yb_global

class cycle:
    def __init__(self,\
                 cbl_settings,\
                 nr,\
                 nens,\
                 FILTER,\
                 trun,\
                 assimilation_interval,\
                 ocoord,\
                 observations,\
                 obs_error_sdev_assimilate,\
                 localization_cutoff,\
                 inflate_rtps,\
                 inflation_rtps_alpha,\
                 rnseed):

        # How many cycles do you need to run?
        ncycles = int(trun/assimilation_interval)
        time = (np.arange(ncycles)+1)*assimilation_interval

        # Define state vector length
        state_vector_length = nr.x0.size
        initial_state_vector = nr.x0
        xcoord = nr.state_coordinates
        nobs = observations.shape[1]
        npar = cbl_settings['parameter_number']

        # Special case: if you re-use a nature run in a experiment without
        # parameter estimation, the extra elements in the state vector must
        # be stripped, otherwise the ensemble members won't be initialized
        # properly
        if hasattr(nr,"has_parameters_in_state") and cbl_settings['do_parameter_estimation'] == False:
            print("Warning: you are re-using a nature run with augmented state in a run without parameter estimation")
            state_vector_length -= nr.parameter_number
            initial_state_vector = np.delete(nr.x0,-nr.parameter_number,0)
            xcoord = np.delete(nr.state_coordinates,-nr.parameter_number)

        # Preallocate arrays
        initial_state     = np.zeros((1,state_vector_length,nens)) + np.nan
        backgrounds       = np.zeros((ncycles,state_vector_length,nens)) + np.nan
        analyses          = np.zeros((ncycles,state_vector_length,nens)) + np.nan
        model_equivalents = np.zeros((ncycles,nobs,nens)) + np.nan
        if cbl_settings['do_parameter_estimation'] and cbl_settings['return_covariances_increments_and_innovations']:
            increments = np.zeros((ncycles,nobs,npar)) + np.nan
            cov_py = np.zeros((ncycles,nobs,npar)) + np.nan
            cov_pp = np.zeros((ncycles,nobs,npar)) + np.nan
            cov_yy = np.zeros((ncycles,nobs)) + np.nan
            innov = np.zeros((ncycles,nobs)) + np.nan

        # Turn inflation coefficients into an array
        inflation_coefficients_rtps = np.ones(state_vector_length)*inflation_rtps_alpha

        # Initialize ensemble
        # Get the right model class from the nature run.
        # The class must have "initialize", "run" and "update" methods.
        if isinstance(nr,CBL):
            # Initialize assimilation ensemble by adding
            # random perturbations to initial state of nature run.
            da = CBL(cbl_settings)
            x0 = np.tile(initial_state_vector,(nens,1)).T
            da.initialize(x0,coordinates=xcoord)
            # In case of parameter estimation, update variance inflation settings
            if da.do_parameter_estimation:
                inflation_coefficients_rtps[-da.parameter_number:] = da.parameter_inflation_rtps_alpha

        # Store initial ensemble state
        initial_state[0,:,:] = da.x0[:,:]

        history = []
        # Start cycling
        for k in range(ncycles):
            if verbose: print('Cycle %04u    : '%k,end='')

            # Compute distances between current observations and state vector variables
            dist = compute_distances(xcoord,ocoord[k,:])

            # Replace any NaN distances with 0
            # NaN distances only occur when the state is augmented with global
            # parameters, whose location is undefined. Setting the distance to
            # global parameters equal to zero ensures that the covariance
            # matrix for global parameters is identity, as required by
            # global updating.
            np.nan_to_num(dist,copy=False)

            # Integrate up to analysis time
            da.maxtime = assimilation_interval
            output_full_history = False
            if output_full_history:
                da.run(output_full_history=True)
                history.append(da.history)
            else:
                da.run()

            # Store prior
            backgrounds[k,:,:] = da.x[:,:]

            # If you want to add/simulate some error growth in the run, before the analysis,
            # perturb the prior state here.
            if isinstance(da,CBL):
                if cbl_settings["simulate_error_growth"]:
                    backgrounds[k,:,:] = CBL.perturb_state(da,backgrounds[k,:,:])

            # Compute prior spread (necessary for variance inflation)
            if inflate_rtps:
                sigp = np.std(backgrounds[k,:,:],axis=1)

            # Assimilate
            if FILTER == 'EAKF':
                if cbl_settings['do_parameter_estimation'] and cbl_settings['return_covariances_increments_and_innovations']:
                    updates,model_equivalents[k,:,:],cov_xx,cov_xy,cov_yy_dum,innov_dum,increm_dum = eakf(backgrounds[k,:,:],observations[k,:],
                                obs_error_sdev_assimilate,
                                xcoord,
                                ocoord[k,:],
                                dist,
                                localization_cutoff=localization_cutoff,return_covariances_increments_and_innovations=True)
                    increments[k,:,-npar] = increm_dum[:,-npar]
                    cov_pp[k,:,-npar] = cov_xx[:,-npar]
                    cov_py[k,:,-npar] = cov_xy[:,-npar]
                    cov_yy[k,:] = cov_yy_dum
                    innov[k,:] = innov_dum
                else:
                    updates,model_equivalents[k,:,:] = eakf(backgrounds[k,:,:],observations[k,:],
                                obs_error_sdev_assimilate,
                                xcoord,
                                ocoord[k,:],
                                dist,
                                localization_cutoff=localization_cutoff)
            elif FILTER == 'LETKF':
                updates,model_equivalents[k,:,:] = letkf(backgrounds[k,:,:],observations[k,:],
                            obs_error_sdev_assimilate,
                            xcoord,
                            ocoord[k,:],
                            dist,
                            localization_cutoff=localization_cutoff,
                            inflation_rho=1.0)

            x0 = updates

            # Posterior variance inflation (relaxation to prior spread)
            # http://dx.doi.org/10.1175/MWR-D-11-00276.1
            # Variance inflation is optional for regular state vector elements
            if inflate_rtps:
                # Add 1e-10 to posterior variance (denominator) to avoid NaN
                sigu = np.std(x0,axis=1)
                inflation_factor = np.tile(inflation_coefficients_rtps*(sigp-sigu)/(sigu+1e-10),(nens,1))
                xbm = np.tile(x0.mean(axis=1),(nens,1))
                xbp = x0.T-xbm
                x0 = (xbm+xbp*(1.+inflation_factor)).T

            # Store posterior
            analyses[k,:,:] = x0

            # Save initial conditions for next cycle
            da.update(x0)

        # Store the output
        self.ncycles = ncycles
        self.time = time
        self.initial_state = initial_state
        self.backgrounds = backgrounds
        self.analyses = analyses
        self.obs_coordinates = ocoord
        self.observations = observations
        self.model_equivalents = model_equivalents
        self.state_coordinates = xcoord
        self.obs_error_sdev_assimilate = obs_error_sdev_assimilate
        self.history = history
        self.nz = nr.nz
        self.zt = nr.zt
        self.nens = nens
        self.initial_perturbed_parameters = da.initial_perturbed_parameters
        if cbl_settings['do_parameter_estimation'] and cbl_settings['return_covariances_increments_and_innovations']:
            self.cov_py = cov_py
            self.cov_pp = cov_pp
            self.cov_yy = cov_yy
            self.innov = innov
            self.increments = increments

"""
class experiment:
    def __init__(self,settings):
        # Read attributes from dictionary of settings
        for k, v in settings.items():
            setattr(self, k, v)        
        if 0 < self.inflation_rtps_alpha <= 1:
            self.inflate_rtps = True
        else:
            self.inflate_rtps = False

        # Consistency check on observation info
        nobs = self.obs_coordinates.size
        assert (self.obs_coordinates.size==self.obs_kinds.size==self.obs_error_sdev_assimilate.size==self.obs_error_sdev_generate.size),\
            f'obs_coordinates, obs_kinds and obs_error_sdev must have the same size,\
            got {self.obs_coordinates.size,self.obs_kinds.size,self.obs_error_sdev_assimilate.size,self.obs_error_sdev_assimilate.size}'

        # Consistency check on time coordinates
        nassim = self.trun/self.assimilation_interval
        assert (nassim.is_integer()),\
            f'trun/assimilation_interval must be integer,\
            got {self.trun,self.assimilation_interval,self.trun/self.assimilation_interval}'

        # If you don't have a pre-existing nature run, you need to run it
        if not hasattr(self,'nr'):
            # Define initial conditions of nature run
            if self.tspinup > 0:
                # Create initial state with random perturbations, then integrate
                # one deterministic run for the spinup period
                # Initial condition for the nature run is the last state of the spinup run
                if verbose: print('Spinup        : ',end='')
                sp = CBL(self.cbl_settings)
                sp.maxtime = self.tspinup
                sp.initialize(1) 
                sp.run()
                initial_conditions = sp.x
                coords = sp.state_coordinates
            else:
                # No spinup, nature run gets unperturbed initial condition
                sp = CBL(self.cbl_settings)
                sp.initialize(1)
                initial_conditions = sp.x0
                coords = sp.state_coordinates

            # Nature run: initialize, then integrate
            if verbose: print('Nature run    : ',end='')
            nr = CBL(self.cbl_settings)
            nr.maxtime = self.trun
            nr.initialize(initial_conditions,coordinates=coords)
            nr.run(output_full_history=True)

            # Draw observations from nature run
            # Need nested loops because observation times (k) are independent
            # and observation error variance depends on location (kk)
            state_coordinates = nr.zt
            truths = np.zeros((int(nassim),nobs))+np.nan
            observations = np.zeros((int(nassim),nobs))+np.nan
            for i in range(int(nassim)):
                valid_time = (i+1)*self.assimilation_interval
                assert(np.mod(self.assimilation_interval,nr.dt)==0), f'Assimilation interval must be a multiple of dt, got {self.assimilation_interval,nr.dt}'
                time_index = np.argwhere(nr.history['time'] == valid_time)[0][0]
                for j in range(nobs):
                    obs_coordinate = self.obs_coordinates[j]
                    if self.obs_kinds[j]=='theta':
                        variable = nr.history['theta'][:,time_index]
                    elif self.obs_kinds[j]=='u':
                        variable = nr.history['u'][:,time_index]
                    elif self.obs_kinds[j]=='v':
                        variable = nr.history['v'][:,time_index]
                    truths[i,j] = observation_operator(variable,state_coordinates,obs_coordinate)
                    # Change the seed every time you go through this,
                    # otherwise the perturbations are always the same;
                    # and you have bias instead of randomness
                    seed = self.rnseed+j*1000+i*100000
                    RNG = np.random.default_rng(seed=seed)
                    observations[i,j] = truths[i,j]+\
                        RNG.normal(0,self.obs_error_sdev_generate[j])

            # Store truths and observations
            self.truths = truths
            self.observations = observations

            # Store nature run
            self.nr = nr

        # Run assimilation cycle
        if not hasattr(self,'da'):
            self.da = cycle(self.cbl_settings,
                        self.nr,
                        self.nens,
                        self.FILTER,                        
                        self.trun,
                        self.assimilation_interval,
                        self.obs_coordinates,
                        self.truths,
                        self.observations,
                        self.obs_error_sdev_assimilate,
                        self.localization_cutoff,
                        self.inflate_rtps,
                        self.inflation_rtps_alpha,
                        self.rnseed)
    
        # Compute diagnostics
        if not hasattr(self,'dg'):
            self.dg = diagnostics(self.da)
"""
class experiment:
    def __init__(self,settings):
        # Read attributes from dictionary of settings
        for k, v in settings.items():
            setattr(self, k, v)        
        if 0 < self.inflation_rtps_alpha <= 1:
            self.inflate_rtps = True
        else:
            self.inflate_rtps = False

        ################################################################################################
        # CASE 1: synthetic observations (OSSE with nature run)
        if not hasattr(self,'nr') and self.type == 'OSSE':
            # Consistency check on observation info
            nobs = self.obs_coordinates.size
            assert (self.obs_coordinates.size==self.obs_kinds.size==self.obs_error_sdev_assimilate.size==self.obs_error_sdev_generate.size),\
                f'obs_coordinates, obs_kinds and obs_error_sdev must have the same size,\
                got {self.obs_coordinates.size,self.obs_kinds.size,self.obs_error_sdev_assimilate.size,self.obs_error_sdev_assimilate.size}'

            # Consistency check on time coordinates
            nassim = self.trun/self.assimilation_interval
            assert (nassim.is_integer()),\
                f'trun/assimilation_interval must be integer,\
                got {self.trun,self.assimilation_interval,self.trun/self.assimilation_interval}'

            # If you don't have a pre-existing nature run, you need to run it

            # Define initial conditions of nature run
            if self.tspinup > 0:
                # Create initial state with random perturbations, then integrate
                # one deterministic run for the spinup period
                # Initial condition for the nature run is the last state of the spinup run
                if verbose: print('Spinup        : ',end='')
                sp = CBL(self.cbl_settings)
                sp.maxtime = self.tspinup
                sp.initialize(1) 
                sp.run()
                initial_conditions = sp.x
                coords = sp.state_coordinates
            else:
                # No spinup, nature run gets unperturbed initial condition
                sp = CBL(self.cbl_settings)
                sp.initialize(1)
                initial_conditions = sp.x0
                coords = sp.state_coordinates

            # Nature run: initialize, then integrate
            if verbose: print('Nature run    : ',end='')
            nr = CBL(self.cbl_settings)
            nr.maxtime = self.trun
            nr.initialize(initial_conditions,coordinates=coords)
            nr.run(output_full_history=True)

            # Draw observations from nature run
            # Need nested loops because observation times (k) are independent
            # and observation error variance depends on location (kk)
            state_coordinates = nr.zt
            truths = np.zeros((int(nassim),nobs))+np.nan
            observations = np.zeros((int(nassim),nobs))+np.nan
            for i in range(int(nassim)):
                valid_time = (i+1)*self.assimilation_interval
                assert(np.mod(self.assimilation_interval,nr.dt)==0), f'Assimilation interval must be a multiple of dt, got {self.assimilation_interval,nr.dt}'
                time_index = np.argwhere(nr.history['time'] == valid_time)[0][0]
                for j in range(nobs):
                    obs_coordinate = self.obs_coordinates[j]
                    if self.obs_kinds[j]=='theta':
                        variable = nr.history['theta'][:,time_index]
                    elif self.obs_kinds[j]=='u':
                        variable = nr.history['u'][:,time_index]
                    elif self.obs_kinds[j]=='v':
                        variable = nr.history['v'][:,time_index]
                    truths[i,j] = observation_operator(variable,state_coordinates,obs_coordinate)
                    # Change the seed every time you go through this,
                    # otherwise the perturbations are always the same;
                    # and you have bias instead of randomness
                    seed = self.rnseed+j*1000+i*100000
                    RNG = np.random.default_rng(seed=seed)
                    observations[i,j] = truths[i,j]+\
                        RNG.normal(0,self.obs_error_sdev_generate[j])

            # In this case the observations coordinates are the same at all analysis times,
            # so a 1-d vector. Expand it to 2d, with the same dimensions
            # as the observations vector
            if self.obs_coordinates.ndim == 1:
                ocoords = self.obs_coordinates[None,:]+np.zeros(observations.shape)

            # Randomize order of observations
            if self.randomize_obs:
                observations,ocoords,truths = random_sort_observations(observations,ocoords,truths)

            # Store nature run, truths and observations
            self.nr = nr
            self.truths = truths
            self.observations = observations
            self.obs_coordinates = ocoords

        ################################################################################################
        # CASE 2: real (or pseudo-) observations
        if not hasattr(self,'nr') and self.type == 'real':

            # Read observations and their coordinates
            observations, ocoords = decode_observations(self.observation_files)
            assert (observations.shape == ocoords.shape),\
                f'observation and coordinate arrays must have the same size'

            # Consistency check on time coordinates
            nassim = self.trun/self.assimilation_interval
            assert (nassim.is_integer()),\
                f'trun/assimilation_interval must be integer,\
                got {self.trun,self.assimilation_interval,self.trun/self.assimilation_interval}'

            # Randomize order of observations
            if self.randomize_obs:
                observations,ocoords = random_sort_observations(observations,ocoords)

            # Throw away any observations in the spinup time before assimilations begin
            # Exclude the initial time too (so the +1 in the line below)
            nspinup = self.tspinup/self.assimilation_interval+1
            observations = observations[int(nspinup):,:]
            ocoords = ocoords[int(nspinup):,:]

            # Throw away any observations beyond tspinup+trun
            observations = observations[:int(nassim),:]
            ocoords = ocoords[:int(nassim),:]

            # There is no nature run and no truth in this case.
            # The only thing you can store is the observations.
            self.observations = observations
            self.obs_coordinates = ocoords

            # The da cycle depends on a "nr" structure to get information
            # about the model properties.
            # Depending on the configuration, the da cycle could 
            # create the initial ensemble perturbations around
            # a spun-up model state, which is stored in nr.x0.
            # So, run the forecast model for the spinup time if necessary.
            # Define initial conditions of nature run
            if self.tspinup > 0:
                # Create initial state with random perturbations, then integrate
                # one deterministic run for the spinup period
                # Initial condition for the nature run is the last state of the spinup run
                if verbose: print('Spinup        : ',end='')
                sp = CBL(self.cbl_settings)
                sp.maxtime = self.tspinup
                sp.initialize(1) 
                sp.run()
                initial_conditions = sp.x
                coords = sp.state_coordinates
            else:
                # No spinup, nature run gets unperturbed initial condition
                sp = CBL(self.cbl_settings)
                sp.initialize(1)
                initial_conditions = sp.x0
                coords = sp.state_coordinates

            # Nature run: initialize, then integrate
            self.nr = CBL(self.cbl_settings)
            self.nr.maxtime = self.trun
            self.nr.initialize(initial_conditions,coordinates=coords)

        ################################################################################################
        # Run assimilation cycle
        if not hasattr(self,'da'):
            self.da = cycle(self.cbl_settings,
                        self.nr,
                        self.nens,
                        self.FILTER,                        
                        self.trun,
                        self.assimilation_interval,
                        self.obs_coordinates,
                        self.observations,
                        self.obs_error_sdev_assimilate,
                        self.localization_cutoff,
                        self.inflate_rtps,
                        self.inflation_rtps_alpha,
                        self.rnseed)

        # If truths (from nature run) exist, add them to the da cycle structure.
        # That's necessary for the diagnostics package to work properly.
        if hasattr(self,'truths'):
            self.da.truths = self.truths

        # Compute diagnostics
        if not hasattr(self,'dg'):
            self.dg = diagnostics(self.da)

class diagnostics:
    def __init__(self,da):
        # ntim: number of verification times in the sample
        # nens: number of ensemble members
        # nvar: number of model state variables
        # nobs: number of observations
        
        # Rename variables for shorthand
        ntim = da.observations.shape[0] 
        nobs = da.observations.shape[1]
        nens = da.backgrounds.shape[2]
        ocoords = da.obs_coordinates   # nobs or ntim*nobs
        xcoords = da.state_coordinates # nvar
        o = da.observations            # ntim*nobs
        if hasattr(da,'truths'):
            t = da.truths              # ntim*nobs
            truth_exists = True
        else:
            truth_exists = False

        # Get observation-space priors from da output
        b = da.model_equivalents # ntim*nobs*nens

        # To get observation-space posteriors (ntim*nvar*nens),
        # apply observation operator to analyses
        a = np.zeros((ntim,nobs,nens))+np.nan

        # In this case, observation coordinates do not change between analysis times
        if ocoords.ndim == 1:
            for i in range(ntim):
                for j in range(nobs):
                    for k in range(nens):
                        a[i,j,k] = observation_operator(da.analyses[i,:,k],xcoords,ocoords[j])

        # In this case, observation coordinates are different at each analysis time
        elif ocoords.ndim == 2:
            for i in range(ntim):
                for j in range(nobs):
                    for k in range(nens):
                        a[i,j,k] = observation_operator(da.analyses[i,:,k],xcoords,ocoords[i,j])

        # Compute ensemble mean and spread
        meana  = a.mean(axis=2) # ntim*nobs
        meanb  = b.mean(axis=2) # ntim*nobs
        sigmaa = a.std(axis=2)  # ntim*nobs
        sigmab = b.std(axis=2)  # ntim*nobs

        # Compute space-averaged analysis RMSE and spread
        aRMSD_x = np.sqrt((meana-o)**2).mean(axis=1) # ntim
        bRMSD_x = np.sqrt((meanb-o)**2).mean(axis=1) # ntim
        if truth_exists:
            aRMSE_x = np.sqrt((meana-t)**2).mean(axis=1) # ntim
            bRMSE_x = np.sqrt((meanb-t)**2).mean(axis=1) # ntim
        aSprd_x = sigmaa.mean(axis=1) # ntim
        bSprd_x = sigmab.mean(axis=1) # ntim

        # Compute time-averaged analysis RMSE and spread
        aRMSD_t = np.sqrt((meana-o)**2).mean(axis=0) # nobs
        bRMSD_t = np.sqrt((meanb-o)**2).mean(axis=0) # nobs
        if truth_exists:
            aRMSE_t = np.sqrt((meana-t)**2).mean(axis=0) # nobs
            bRMSE_t = np.sqrt((meanb-t)**2).mean(axis=0) # nobs
        aSprd_t = sigmaa.mean(axis=0) # nobs
        bSprd_t = sigmab.mean(axis=0) # nobs

        # Compute deviations
        omina = o-meana     # ntim*nobs 
        ominb = o-meanb     # ntim*nobs 
        aminb = meana-meanb # ntim*nobs 

        # Time-averaged Desroziers statistics
        vara_ta = (sigmaa**2).mean(axis=0) # nobs
        varb_ta = (sigmab**2).mean(axis=0) # nobs
        dobdob_ta = (ominb*ominb).mean(axis=0) # nobs
        dabdob_ta = (aminb*ominb).mean(axis=0) # nobs
        doadob_ta = (omina*ominb).mean(axis=0) # nobs
        dabdoa_ta = (aminb*omina).mean(axis=0) # nobs

        # Compute histograms of ensemble mean departures.
        # Histograms are meaningless if you only have 1 observation.
        if o.size > 1:
            nbins = 100
            #
            edges = np.linspace(-omina.std()*4,omina.std()*4,nbins+1)
            hist_omina_global   = np.histogram(omina,bins=edges)[0]/omina.size/(edges[1]-edges[0])
            hist_omina_x = 0.5*(edges[1:]+edges[:-1])
            hist_omina_local = []
            for i in range(nobs):
                hist_omina_local.append(np.histogram(omina[:,i],bins=edges)[0]/omina[:,i].size/(edges[1]-edges[0]))
            #
            edges = np.linspace(-ominb.std()*4,ominb.std()*4,nbins+1)
            hist_ominb_global   = np.histogram(ominb,bins=edges)[0]/ominb.size/(edges[1]-edges[0])
            hist_ominb_x = 0.5*(edges[1:]+edges[:-1])
            hist_ominb_local = []
            for i in range(nobs):
                hist_ominb_local.append(np.histogram(ominb[:,i],bins=edges)[0]/ominb[:,i].size/(edges[1]-edges[0]))
            #
            edges = np.linspace(-aminb.std()*4,aminb.std()*4,nbins+1)
            hist_aminb_global   = np.histogram(aminb,bins=edges)[0]/aminb.size/(edges[1]-edges[0])
            hist_aminb_x = 0.5*(edges[1:]+edges[:-1])
            hist_aminb_local = []
            for i in range(nobs):
                hist_aminb_local.append(np.histogram(aminb[:,i],bins=edges)[0]/aminb[:,i].size/(edges[1]-edges[0]))

        # Analysis & background mean error and spread
        self.aRMSD_x = aRMSD_x # ntim
        self.aSprd_x = aSprd_x # ntim
        self.aRMSD_t = aRMSD_t # nobs
        self.aSprd_t = aSprd_t # nobs
        self.bRMSD_x = bRMSD_x # ntim
        self.bSprd_x = bSprd_x # ntim
        self.bRMSD_t = bRMSD_t # nobs
        self.bSprd_t = bSprd_t # nobs
        if truth_exists:
            self.aRMSE_x = aRMSE_x # ntim
            self.aRMSE_t = aRMSE_t # nobs
            self.bRMSE_x = bRMSE_x # ntim
            self.bRMSE_t = bRMSE_t # nobs

        # All these with dimensions ntim*nobs
        self.meana = meana
        self.meanb = meanb
        self.sigmaa = sigmaa
        self.sigmab = sigmab
        self.omina = omina
        self.ominb = ominb
        self.aminb = aminb

        # All these with dimensions nobs
        self.varo    = da.obs_error_sdev_assimilate**2
        self.vara_ta = vara_ta
        self.varb_ta = varb_ta
        self.dobdob_ta = dobdob_ta
        self.dabdob_ta = dabdob_ta
        self.doadob_ta = doadob_ta
        self.dabdoa_ta = dabdoa_ta
        self.sigma_ominb = ominb.std(axis=0)
        self.mean_ominb  = ominb.mean(axis=0)
        self.sigma_omina = omina.std(axis=0)
        self.mean_omina  = omina.mean(axis=0)

        # All these are scalars
        self.sigma_ominb_global = ominb.std()
        self.mean_ominb_global  = ominb.mean()
        self.sigma_omina_global = omina.std()
        self.mean_omina_global  = omina.mean()

        # Store histograms, if they were computed
        if o.size > 1:
            # All these with dimensions nobs * nbins
            self.hist_omina_local = hist_omina_local
            self.hist_ominb_local = hist_ominb_local
            self.hist_aminb_local = hist_aminb_local

            # All these with dimensions nbins
            self.hist_omina_global = hist_omina_global
            self.hist_omina_x      = hist_omina_x
            self.hist_ominb_global = hist_ominb_global
            self.hist_ominb_x      = hist_ominb_x
            self.hist_aminb_global = hist_aminb_global
            self.hist_aminb_x      = hist_aminb_x