#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as p
import pickle
import os
from copy import deepcopy

# Own modules
from models import CBL
from ENDA import experiment, transform_log, transform_none, transform_pit
from graphics import *

if __name__ == '__main__':

    # Global settings
    nobs = 16
    nens = 200
    sigma_b_init = 0.1 # increased by a factor 10 in experiment B
    sigma_o_gn = 0.0 # set equal to sigma_o_as in experiments C and F
    sigma_o_as = 0.1 # increased by a factor 10 in experiments C and F

    # Define seed for random number generation, to ensure reproducibility
    rnseed = 181612

    # Default CBL model settings
    default_cbl_settings ={
        # Physical parameters
        'g' : 9.806,
        'f' : 1e-4,
        'kvonk' : 0.4,
        'z0' : 0.1,
        'ug' : 0,
        'vg' : 10,
        'theta_0' : 290,
        'gamma' : 3e-3,
        'Hmax' : 0.1,
        'H0_perturbation_ampl_init' : 0.0,
        'H0_perturbation_ampl_time' : 0.0,
        'exct' : 0.3,
        'pfac' : 1.5,
        'Kmin' : 0.1,
        'Kmax' : 200,
        # Model formulation
        'is_bwind' : False,
        'is_cgrad' : True,
        'is_cycle' : False,
        # Domain parameters (Kmax determines dt)
        'dz' : 50,
        'ztop' : 4000,
        # Ensemble generation parameters (only effective for ensemble runs)
        # Part 1: perturbation of initial state
        # perturbation_type can be "smooth", "random", "uniform"
        # the adjective refers to vertical variability
        'rnseed' : rnseed,
        'perturb_ensemble_state' : True,
        'perturbations_type' : "smooth",
        'perturbations_theta_amplitude' : sigma_b_init,
        'perturbations_uv_amplitude' : 0.1,
        'perturbations_smooth_number' : 11,
        'perturbations_symmetric' : True,
        'simulate_error_growth' : False,
        'error_growth_perturbations_amplitude' : 0.0,
        # Part 2: perturbation of parameters
        'perturb_ensemble_parameters' : True,
        'parameter_number' : 1,
        'parameter_transform' : [ transform_pit ],
        'parameter_ensemble_min' : np.array([0.5]),
        'parameter_ensemble_max' : np.array([4.5]),
        'parameter_true' : np.array([1.5]),
        # In case of parameter estimation
        'do_parameter_estimation' : True,
        'parameter_inflation_rtps_alpha' : np.array([0.8]),
        'return_covariances_increments_and_innovations' : True
        }
    integration_dt = 0.25*default_cbl_settings["dz"]**2/default_cbl_settings["Kmax"]

    # Default assimilation settings
    default_da_settings = {'cbl_settings' : default_cbl_settings,
            'tspinup' : 3600,
            'trun' : 21600,
            'assimilation_interval' : 300,
            'randomize_obs': True,
            'obs_coordinates' : np.linspace(0,2000,nobs),
            'obs_kinds' : np.array(['theta']*nobs),
            'obs_error_sdev_generate' : np.ones(nobs)*0,
            'obs_error_sdev_assimilate' : np.ones(nobs)*sigma_o_as,
            'localization_cutoff' : 400,
            'nens' : nens,
            'FILTER' : 'EAKF',
            'inflation_rtps_alpha' : 0.2,
            'rnseed' : rnseed,
            'path' : "./runs/seed=%s_enssize=%s_EAKF_6hrs/"%(rnseed,nens)
            }

    # Safety check on time steps
    assert np.mod(default_da_settings['assimilation_interval'],integration_dt)==0,\
        "Assimilation interval must be an integer multiplier of model dt"

    # Whether or not to run experiments without parameter estimation
    # Applies only to sets B&C; no-PE run is always computed for experiments A&D
    noPE_runs = False

    # Check if output path exists, if not, create it.
    if not os.path.exists(default_da_settings['path']):
        os.makedirs(default_da_settings['path'])

    ############################################################################
    # Experiment A (control)
    # Create a copy of the default settings
    cbl_settings_A = dict(default_cbl_settings)
    da_settings_A = dict(default_da_settings)

    # Run it
    try:
        exp_A = pickle.load(open(default_da_settings['path']+"exp_A.pickle", "rb"))
    except:
        exp_A = experiment(da_settings_A)
        setattr(exp_A,'label','A')
        pickle.dump(exp_A, open(default_da_settings['path']+'exp_A.pickle', 'wb')) 

    # Store the nature run, because it is used also in other experiments/figures
    exp_A.nr.has_parameters_in_state = True
    nature_run = exp_A.nr
    truths = exp_A.truths
    observations = exp_A.observations

    # Corresponding experiment without parameter estimation
    cbl_settings_A_noPE = dict(cbl_settings_A)
    da_settings_A_noPE = dict(da_settings_A)
    da_settings_A_noPE['nr'] = deepcopy(nature_run)
    da_settings_A_noPE['truths'] = truths
    da_settings_A_noPE['observations'] = observations
    da_settings_A_noPE['nr'].do_parameter_estimation = False
    cbl_settings_A_noPE['do_parameter_estimation'] = False
    cbl_settings_A_noPE['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    da_settings_A_noPE['cbl_settings'] = cbl_settings_A_noPE

    # Run it
    try:
        exp_A_noPE = pickle.load(open(default_da_settings['path']+'exp_A_noPE.pickle', "rb"))
    except:
        exp_A_noPE = experiment(da_settings_A_noPE)
        pickle.dump(exp_A_noPE, open(default_da_settings['path']+'exp_A_noPE.pickle', 'wb')) 

    ########################################################################
    # Experiment B1
    # Create a copy of the default settings
    # Then re-use the available nature run and derived information
    cbl_settings_B1 = dict(default_cbl_settings)
    da_settings_B1 = dict(default_da_settings)
    da_settings_B1['nr'] = deepcopy(nature_run)
    da_settings_B1['truths'] = truths
    da_settings_B1['observations'] = observations

    # Change settings as necessary
    cbl_settings_B1['perturbations_theta_amplitude'] = sigma_b_init*10
    cbl_settings_B1['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    da_settings_B1['cbl_settings'] = cbl_settings_B1
    da_settings_B1['obs_error_sdev_assimilate'] = np.ones(nobs)*sigma_o_as

    # Run and save to disk
    try:
        exp_B1 = pickle.load(open(default_da_settings['path']+'exp_B1.pickle', "rb"))
    except:
        exp_B1 = experiment(da_settings_B1)
        setattr(exp_B1,'label','B1')
        pickle.dump(exp_B1, open(default_da_settings['path']+'exp_B1.pickle', 'wb')) 

    if noPE_runs:
        # Corresponding experiment without parameter estimation
        cbl_settings_B1_noPE = dict(cbl_settings_B1)
        da_settings_B1_noPE = dict(da_settings_B1)
        cbl_settings_B1_noPE['do_parameter_estimation'] = False
        da_settings_B1_noPE['nr'] = deepcopy(nature_run)
        da_settings_B1_noPE['nr'].do_parameter_estimation = False
        da_settings_B1_noPE['cbl_settings'] = cbl_settings_B1_noPE

        # Run and save to disk
        try:
            exp_B1_noPE = pickle.load(open(default_da_settings['path']+'exp_B1_noPE.pickle', "rb"))
        except:
            exp_B1_noPE = experiment(da_settings_B1_noPE)
            pickle.dump(exp_B1_noPE, open(default_da_settings['path']+'exp_B1_noPE.pickle', 'wb'))

    ########################################################################
    # Experiment B2
    # Create a copy of the default settings
    # Then re-use the available nature run and derived information
    cbl_settings_B2 = dict(default_cbl_settings)
    da_settings_B2 = dict(default_da_settings)
    da_settings_B2['nr'] = deepcopy(nature_run)
    da_settings_B2['truths'] = truths
    da_settings_B2['observations'] = observations

    # Change settings as necessary
    cbl_settings_B2['perturbations_theta_amplitude'] = sigma_b_init
    cbl_settings_B2['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    da_settings_B2['cbl_settings'] = cbl_settings_B2
    da_settings_B2['obs_error_sdev_assimilate'] = np.ones(nobs)*sigma_o_as*10

    # Run and save to disk
    try:
        exp_B2 = pickle.load(open(default_da_settings['path']+'exp_B2.pickle', "rb"))
    except:
        exp_B2 = experiment(da_settings_B2)
        setattr(exp_B2,'label','B2')
        pickle.dump(exp_B2, open(default_da_settings['path']+'exp_B2.pickle', 'wb')) 

    if noPE_runs:
        # Corresponding experiment without parameter estimation
        cbl_settings_B2_noPE = dict(cbl_settings_B2)
        da_settings_B2_noPE = dict(da_settings_B2)
        cbl_settings_B2_noPE['do_parameter_estimation'] = False
        da_settings_B2_noPE['nr'] = deepcopy(nature_run)
        da_settings_B2_noPE['nr'].do_parameter_estimation = False
        da_settings_B2_noPE['cbl_settings'] = cbl_settings_B2_noPE

        # Run and save to disk
        try:
            exp_B2_noPE = pickle.load(open(default_da_settings['path']+'exp_B2_noPE.pickle', "rb"))
        except:
            exp_B2_noPE = experiment(da_settings_B2_noPE)
            pickle.dump(exp_B2_noPE, open(default_da_settings['path']+'exp_B2_noPE.pickle', 'wb'))

    ########################################################################
    # Experiment B3
    # Create a copy of the default settings
    # Then re-use the available nature run and derived information
    cbl_settings_B3 = dict(default_cbl_settings)
    da_settings_B3 = dict(default_da_settings)
    da_settings_B3['nr'] = deepcopy(nature_run)
    da_settings_B3['truths'] = truths
    da_settings_B3['observations'] = observations

    # Change settings as necessary
    cbl_settings_B3['perturbations_theta_amplitude'] = sigma_b_init*10
    cbl_settings_B3['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    da_settings_B3['cbl_settings'] = cbl_settings_B3
    da_settings_B3['obs_error_sdev_assimilate'] = np.ones(nobs)*sigma_o_as*10

    # Run and save to disk
    try:
        exp_B3 = pickle.load(open(default_da_settings['path']+'exp_B3.pickle', "rb"))
    except:
        exp_B3 = experiment(da_settings_B3)
        setattr(exp_B3,'label','B3')
        pickle.dump(exp_B3, open(default_da_settings['path']+'exp_B3.pickle', 'wb')) 

    if noPE_runs:
        # Corresponding experiment without parameter estimation
        cbl_settings_B3_noPE = dict(cbl_settings_B3)
        da_settings_B3_noPE = dict(da_settings_B3)
        cbl_settings_B3_noPE['do_parameter_estimation'] = False
        da_settings_B3_noPE['nr'] = deepcopy(nature_run)
        da_settings_B3_noPE['nr'].do_parameter_estimation = False
        da_settings_B3_noPE['cbl_settings'] = cbl_settings_B3_noPE

        # Run and save to disk
        try:
            exp_B3_noPE = pickle.load(open(default_da_settings['path']+'exp_B3_noPE.pickle', "rb"))
        except:
            exp_B3_noPE = experiment(da_settings_B3_noPE)
            pickle.dump(exp_B3_noPE, open(default_da_settings['path']+'exp_B3_noPE.pickle', 'wb'))

    ########################################################################
    # Experiment B4
    # Create a copy of the default settings
    cbl_settings_B4 = dict(default_cbl_settings)
    da_settings_B4 = dict(default_da_settings)

    # Change settings as necessary
    # Changes include generation of observations, so the existing nature run
    # can't be reused.
    cbl_settings_B4['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    da_settings_B4['cbl_settings'] = cbl_settings_B4
    da_settings_B4['obs_error_sdev_generate'] = np.ones(nobs)*sigma_o_as*5
    da_settings_B4['obs_error_sdev_assimilate'] = np.ones(nobs)*sigma_o_as*10

    # Run and save to disk
    try:
        exp_B4 = pickle.load(open(default_da_settings['path']+'exp_B4.pickle', "rb"))
    except:
        exp_B4 = experiment(da_settings_B4)
        setattr(exp_B4,'label','B4')
        pickle.dump(exp_B4, open(default_da_settings['path']+'exp_B4.pickle', 'wb')) 

    if noPE_runs:
        # Corresponding experiment without parameter estimation
        cbl_settings_B4_noPE = dict(cbl_settings_B4)
        da_settings_B4_noPE = dict(da_settings_B4)
        cbl_settings_B4_noPE['do_parameter_estimation'] = False
        da_settings_B4_noPE['cbl_settings'] = cbl_settings_B4_noPE

        # Run and save to disk
        try:
            exp_B4_noPE = pickle.load(open(default_da_settings['path']+'exp_B4_noPE.pickle', "rb"))
        except:
            exp_B4_noPE = experiment(da_settings_B4_noPE)
            pickle.dump(exp_B4_noPE, open(default_da_settings['path']+'exp_B4_noPE.pickle', 'wb')) 

    ########################################################################
    # Experiment C1
    # Create a copy of the default settings
    cbl_settings_C1 = dict(default_cbl_settings)
    da_settings_C1 = dict(default_da_settings)

    # Re-use the available nature run and derived information
    da_settings_C1['nr'] = deepcopy(nature_run)
    da_settings_C1['truths'] = truths
    da_settings_C1['observations'] = observations

    # Change settings as necessary
    cbl_settings_C1['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    cbl_settings_C1['is_cgrad'] = False
    da_settings_C1['cbl_settings'] = cbl_settings_C1

    # Run and save to disk
    try:
        exp_C1 = pickle.load(open(default_da_settings['path']+'exp_C1.pickle', "rb"))
    except:
        exp_C1 = experiment(da_settings_C1)
        setattr(exp_C1,'label','C1')
        pickle.dump(exp_C1, open(default_da_settings['path']+'exp_C1.pickle', 'wb')) 

    if noPE_runs:
        # Corresponding experiment without parameter estimation
        cbl_settings_C1_noPE = dict(cbl_settings_C1)
        da_settings_C1_noPE = dict(da_settings_C1)
        cbl_settings_C1_noPE['do_parameter_estimation'] = False
        da_settings_C1_noPE['cbl_settings'] = cbl_settings_C1_noPE

        # Run and save to disk
        try:
            exp_C1_noPE = pickle.load(open(default_da_settings['path']+'exp_C1_noPE.pickle', "rb"))
        except:
            exp_C1_noPE = experiment(da_settings_C1_noPE)
            pickle.dump(exp_C1_noPE, open(default_da_settings['path']+'exp_C1_noPE.pickle', 'wb')) 

    ########################################################################
    # Experiment C2
    # Create a copy of the default settings
    cbl_settings_C2 = dict(default_cbl_settings)
    da_settings_C2 = dict(default_da_settings)

    # Re-use the available nature run and derived information
    da_settings_C2['nr'] = deepcopy(nature_run)
    da_settings_C2['truths'] = truths
    da_settings_C2['observations'] = observations

    # Change settings as necessary
    cbl_settings_C2['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    cbl_settings_C2['Hmax'] = 0.15
    da_settings_C2['cbl_settings'] = cbl_settings_C2

    # Run and save to disk
    try:
        exp_C2 = pickle.load(open(default_da_settings['path']+'exp_C2.pickle', "rb"))
    except:
        exp_C2 = experiment(da_settings_C2)
        setattr(exp_C2,'label','C2')
        pickle.dump(exp_C2, open(default_da_settings['path']+'exp_C2.pickle', 'wb')) 

    if noPE_runs:
        # Corresponding experiment without parameter estimation
        cbl_settings_C2_noPE = dict(cbl_settings_C2)
        da_settings_C2_noPE = dict(da_settings_C2)
        cbl_settings_C2_noPE['do_parameter_estimation'] = False
        da_settings_C2_noPE['cbl_settings'] = cbl_settings_C2_noPE

        # Run and save to disk
        try:
            exp_C2_noPE = pickle.load(open(default_da_settings['path']+'exp_C2_noPE.pickle', "rb"))
        except:
            exp_C2_noPE = experiment(da_settings_C2_noPE)
            pickle.dump(exp_C2_noPE, open(default_da_settings['path']+'exp_C2_noPE.pickle', 'wb')) 

    ########################################################################
    # Experiment D
    # Create a copy of the default settings
    cbl_settings_D = dict(default_cbl_settings)
    da_settings_D = dict(default_da_settings)

    # Change settings as necessary
    # Changes include generation of observations, so the existing nature run
    # can't be reused.
    cbl_settings_D['initial_perturbed_parameters'] = exp_A.da.initial_perturbed_parameters
    cbl_settings_D['perturbations_theta_amplitude'] = sigma_b_init*10
    cbl_settings_D['Hmax'] = 0.15
    cbl_settings_D['is_cgrad'] = False
    cbl_settings_D['simulate_error_growth'] = True
    cbl_settings_D['error_growth_perturbations_amplitude'] = sigma_b_init*5
    da_settings_D['cbl_settings'] = cbl_settings_D
    da_settings_D['obs_error_sdev_generate'] = np.ones(nobs)*sigma_o_as*5
    da_settings_D['obs_error_sdev_assimilate'] = np.ones(nobs)*sigma_o_as*10

    # Run and save to disk
    try:
        exp_D = pickle.load(open(default_da_settings['path']+'exp_D.pickle', "rb"))
    except:
        exp_D = experiment(da_settings_D)
        setattr(exp_D,'label','D')
        pickle.dump(exp_D, open(default_da_settings['path']+'exp_D.pickle', 'wb')) 

    # Experiment matching D, but without parameter estimation
    if noPE_runs:
        cbl_settings_D_noPE = dict(cbl_settings_D)
        da_settings_D_noPE = dict(da_settings_D)
        cbl_settings_D_noPE['do_parameter_estimation'] = False
        da_settings_D_noPE['cbl_settings'] = cbl_settings_D_noPE

        try:
            exp_D_noPE = pickle.load(open(default_da_settings['path']+'exp_D_noPE.pickle', "rb"))
        except:
            exp_D_noPE = experiment(da_settings_D_noPE)
            pickle.dump(exp_A_noPE, open(default_da_settings['path']+'exp_D_noPE.pickle', 'wb')) 

    ########################################################################

    # Decide what figures to plot
    fig101 = False
    fig102 = False
    fig103 = False
    fig104 = False
    fig105 = False
    fig106 = False
    fig107 = False
    fig108 = False
    fig109 = False
    
    # Other possible, optional, plots
    opt01 = False # assimilation of a single observation
    opt02 = False # assimilation of profiles at two times
    opt03 = False  # assimilation of real observations
    opt04 = True  # assimilation of real observations, different spinup times

    if fig101:

        # Create a copy of the default settings
        cbl_settings = dict(default_cbl_settings)
        maxtime = 21600
        cbl_settings['maxtime'] = maxtime

        # Disable parameter estimation (not used here)
        cbl_settings['do_parameter_estimation'] = False
    
        # Panel a) deterministic run using default settings
        cbl_det = CBL(cbl_settings)
        cbl_det.initialize(1)
        cbl_det.run(output_full_history=True)

        # Panel b) sensitivity to p; use a smaller domain and higher
        # resolution for better-looking plots.
        zmax = 2000
        p_factors = [0.5,1.5,4.5]
        cbl_settings['dz'] = 25
        cbl_settings['ztop'] = zmax

        theta_profiles = []
        for pfac in p_factors:
            cbl_settings['pfac'] = pfac
            cbl_pf = CBL(cbl_settings)
            cbl_pf.initialize(1)
            cbl_pf.run()
            theta_profiles.append(cbl_pf.x[:cbl_pf.nz])

        # Panel c) spread induced by p
        # Do a free ensemble run (ensemble size set expliclity)
        cbl_settings_free = dict(default_cbl_settings)
        cbl_settings_free['maxtime'] = maxtime
        cbl_settings_free['perturb_ensemble_state'] = False
        cbl_free = CBL(cbl_settings_free)
        cbl_free.initialize(nens)
        cbl_free.run(output_full_history=True)

        # Make plots
        ncont = 13
        fig, [[ax4, ax2],[ax1, ax3]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(6,6)

        c1 = ax1.pcolormesh(cbl_det.history['time']/3600,cbl_det.zt,cbl_det.history['theta'],vmin=290,vmax=296)
        ax1.set_ylim([0,zmax])
        ax1.set_ylabel(r'Height (m)')
        ax1.set_xlabel(r'Time (h)')
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        ax1.set_title(r'c) $\overline{\theta}$ (K)')
        p.colorbar(c1,orientation='horizontal')
        ax1.contour(cbl_det.history['time']/3600,cbl_det.zt,cbl_det.history['theta'],
                    np.linspace(cbl_det.theta_0,cbl_det.theta_0+cbl_det.gamma*zmax,ncont),
                    colors='white',
                    linestyles='--',
                    linewidths=0.75)

        ax2 = plot_p(p_factors,theta_profiles,cbl_pf.zt,None,ax=ax2)
        ax2.set_ylabel(r'Height (m)')
        ax2.set_xlabel(r'$\overline{\theta}$ (K)')
        ax2.set_xlim([291,297])
        ax2.set_ylim([0,zmax])
        ax2.legend(loc=4,frameon=False)
        ax2.set_title(r'b)  $\overline{\theta}$ sensitivity to $p$')

        ax3,c3 = plot_spread(cbl_free,ax=ax3)
        ax3.set_ylabel(r'Height (m)')
        ax3.set_title(r'd) $\sigma_\theta$ (K)')
        ax3.set_xlabel('Time (h)')
        ax3.set_xticks(np.arange((maxtime+1)/3600))
        p.colorbar(c3,orientation='horizontal')

        zoverh = np.linspace(0,1,101)
        for pfac in p_factors:
            Koverkws = zoverh*(1-zoverh)**pfac
            ax4.plot(Koverkws,zoverh,label='$p=%4.1f$'%pfac)
        ax4.set_title(r'a) $K_h$ sensitivity to $p$')
        ax4.set_xlabel('$K_h/(\kappa w_s h)$')
        ax4.set_ylabel('$z/h$')
        ax4.set_xlim([0,0.5])
        ax4.legend(loc=4,frameon=False)

        #p.setp(ax2.get_yticklabels(), visible=False)
        #p.setp(ax3.get_yticklabels(), visible=False)
        fig.savefig('fig01.png',format='png',dpi=300)
        p.close(fig)

    if fig102:

        exp_A = pickle.load(open(default_da_settings['path']+"exp_A.pickle", "rb"))

        fig, [[ax0, ax1], [ax2, ax3]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(6,6)
        #
        [ax0,ax1,ax2],c0,c1,c2 = plot_CBL_identifiability(exp_A,da_settings_A['obs_error_sdev_assimilate'][0],ax=[ax0,ax1,ax2])
        ax0.set_title(r'a) Exp. A$_1$, $\rho(p\prime\prime,y_b}$)')
        ax0.set_xlabel('Time (h)')
        ax0.set_ylabel('Height (m)')
        ax1.set_title(r'b) Exp. A$_1$, $\delta\overline{y}\cdot(\sigma_{p\prime\prime}/\sigma_{y^b})$')
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel('Height (m)')
        ax2.set_title(r'c) Exp. A$_1$, $\delta\overline{p}\prime\prime$')
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel('Height (m)')
        ax3 = plot_CBL_PE(exp_A,None,ax=ax3)
        ax3.set_title(r'd) Exp. A$_1$, evolution of $p$')
        ax3.set_xlabel('Time (h)')
        ax3.set_yticks([0,1,2,3,4,5])
        maxtime = exp_A.trun
        ax0.set_xlim([0,maxtime/3600])
        ax1.set_xlim([0,maxtime/3600])
        ax2.set_xlim([0,maxtime/3600])
        ax3.set_xlim([0,maxtime/3600])
        ax0.set_xticks(np.arange((maxtime+1)/3600))
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        ax2.set_xticks(np.arange((maxtime+1)/3600))
        ax3.set_xticks(np.arange((maxtime+1)/3600))
        p.colorbar(c0,orientation='horizontal')
        p.colorbar(c1,orientation='horizontal')
        p.colorbar(c2,orientation='horizontal')
        #
        fig.savefig('fig02.png',format='png',dpi=300)
        p.close(fig)

    if fig103:

        exp_A = pickle.load(open(default_da_settings['path']+"exp_A.pickle", "rb"))
        exp_A_noPE = pickle.load(open(default_da_settings['path']+"exp_A_noPE.pickle", "rb"))

        experiments_pe = [exp_A]
        experiments_nope = [exp_A_noPE]
        labels = ["Exp. A"]

        plot_diagnostics(experiments_pe,experiments_nope,labels,'fig03.png')

    if fig104:

        exp_B1 = pickle.load(open(default_da_settings['path']+"exp_B1.pickle", "rb"))
        exp_B2 = pickle.load(open(default_da_settings['path']+"exp_B2.pickle", "rb"))
        exp_B3 = pickle.load(open(default_da_settings['path']+"exp_B3.pickle", "rb"))
        exp_B4 = pickle.load(open(default_da_settings['path']+"exp_B4.pickle", "rb"))

        fig, [[ax1, ax2], [ax3, ax4]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(6,4)
        #
        ax1 = plot_CBL_PE(exp_B1,None,ax=ax1)
        ax1.set_title(r'a) Exp. B$_1$')
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel(r'$p$')
        #
        ax2 = plot_CBL_PE(exp_B2,None,ax=ax2)
        ax2.set_title(r'b) Exp. B$_2$')
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel(r'$p$')
        #
        ax3 = plot_CBL_PE(exp_B3,None,ax=ax3)
        ax3.set_title(r'c) Exp. B$_3$')
        ax3.set_xlabel('Time (h)')
        ax3.set_ylabel(r'$p$')
        #
        ax4 = plot_CBL_PE(exp_B4,None,ax=ax4)
        ax4.set_title(r'd) Exp. B$_4$')
        ax4.set_xlabel('Time (h)')
        ax4.set_ylabel(r'$p$')
        #
        ax1.set_yticks([0,1,2,3,4,5])
        ax2.sharey(ax1)
        ax3.set_yticks([0,1,2,3,4,5])
        ax4.sharey(ax3)
        #
        fig.savefig('fig04.png',format='png',dpi=300)
        p.close(fig)

    if fig105:

        exp_C1 = pickle.load(open(default_da_settings['path']+"exp_C1.pickle", "rb"))
        exp_C2 = pickle.load(open(default_da_settings['path']+"exp_C2.pickle", "rb"))

        fig, [ax1, ax2] = p.subplots(1,2,constrained_layout=True)
        fig.set_size_inches(6,2)
        #
        ax1 = plot_CBL_PE(exp_C1,None,ax=ax1)
        ax1.set_title(r'a) Exp. C$_1$')
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel(r'$p$')
        #
        ax2 = plot_CBL_PE(exp_C2,None,ax=ax2)
        ax2.set_title(r'b) Exp. C$_2$')
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel(r'$p$')
        #
        ax1.set_yticks([0,1,2,3,4,5])
        ax2.sharey(ax1)
        #
        fig.savefig('fig05.png',format='png',dpi=300)
        p.close(fig)

    if fig106:

        exp_C1 = pickle.load(open(default_da_settings['path']+"exp_C1.pickle", "rb"))
        exp_C2 = pickle.load(open(default_da_settings['path']+"exp_C2.pickle", "rb"))

        npar2 = exp_C1.nr.parameter_number
        npar3 = exp_C2.nr.parameter_number
        updates_C1 = exp_C1.da.analyses[:,:-npar2,:].mean(axis=2).T - exp_C1.da.backgrounds[:,:-npar2,:].mean(axis=2).T
        updates_C2 = exp_C2.da.analyses[:,:-npar3,:].mean(axis=2).T - exp_C2.da.backgrounds[:,:-npar3,:].mean(axis=2).T

        fig, [ax1, ax2] = p.subplots(1,2,constrained_layout=True)
        fig.set_size_inches(6,3)
        c1 = ax1.pcolormesh(exp_C1.da.time/3600,exp_C1.nr.zt,updates_C1,vmin=-0.03,vmax=0.03,cmap='RdBu_r')
        c2 = ax2.pcolormesh(exp_C2.da.time/3600,exp_C2.nr.zt,updates_C2,vmin=-0.03,vmax=0.03,cmap='RdBu_r')
        #
        ax1.set_ylim([0,2000])
        ax2.set_ylim([0,2000])
        ax1.set_title(r'a) Exp. C$_1$, $\overline{\theta}_a-\overline{\theta}_b$')
        ax2.set_title(r'b) Exp. C$_2$, $\overline{\theta}_a-\overline{\theta}_b$')
        ax1.set_xlabel('Time (h)')
        ax2.set_xlabel('Time (h)')
        ax1.set_ylabel('Height (m)')
        ax2.sharey(ax1)
        maxtime = exp_C1.trun
        ax1.set_xlim([0,maxtime/3600])
        ax2.set_xlim([0,maxtime/3600])
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        ax2.set_xticks(np.arange((maxtime+1)/3600))
        p.setp(ax2.get_yticklabels(), visible=False)
        #
        p.colorbar(c1,orientation='horizontal')
        p.colorbar(c2,orientation='horizontal')
        #
        fig.savefig('fig06.png',format='png',dpi=300)
        p.close(fig)

    if fig107:

        exp_D = pickle.load(open(default_da_settings['path']+"exp_D.pickle", "rb"))

        fig, [ax0, ax1] = p.subplots(1,2,constrained_layout=True)
        fig.set_size_inches(6,3)
        #
        ax0,c0 = plot_CBL_identifiability(exp_D,da_settings_D['obs_error_sdev_assimilate'][0],ax=ax0)
        ax0.set_title(r'a) Exp. D, $\rho(p\prime\prime,y_b}$)')
        ax0.set_xlabel('Time (h)')
        ax0.set_ylabel('Height (m)')
        ax1 = plot_CBL_PE(exp_D,None,ax=ax1)
        ax1.set_title(r'b) Exp. D, evolution of $p$')
        ax1.set_xlabel('Time (h)')
        ax1.set_yticks([0,1,2,3,4,5])
        maxtime = exp_D.trun
        ax0.set_xlim([0,maxtime/3600])
        ax1.set_xlim([0,maxtime/3600])
        ax0.set_xticks(np.arange((maxtime+1)/3600))
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        p.colorbar(c0,orientation='horizontal')
        #
        fig.savefig('fig07.png',format='png',dpi=300)
        p.close(fig)

    if fig108:

        exp_A = pickle.load(open(default_da_settings['path']+"exp_A.pickle", "rb"))
        exp_D = pickle.load(open(default_da_settings['path']+"exp_D.pickle", "rb"))

        fig, [[ax0, ax1],[ax2,ax3]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(6,6)
        #
        [ax1,ax0],c0,c1 = plot_CBL_identifiability(exp_A,da_settings_A['obs_error_sdev_assimilate'][0],ax=[ax1,ax0])
        ax0.set_title(r'a) Exp. A$_1$, $\delta\overline{y}$ (K)')
        ax0.set_xlabel('Time (h)')
        ax0.set_ylabel('Height (m)')
        ax1.set_title(r'b) Exp. A$_1$, $\sigma_{p\prime\prime}/\sigma_{y^b}$ (K$^{-1}$)')
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel('Height (m)')
        maxtime = exp_A.trun
        ax0.set_xlim([0,maxtime/3600])
        ax1.set_xlim([0,maxtime/3600])
        ax0.set_xticks(np.arange((maxtime+1)/3600))
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        [ax3,ax2],c2,c3 = plot_CBL_identifiability(exp_D,da_settings_D['obs_error_sdev_assimilate'][0],ax=[ax3,ax2])
        ax2.set_title(r'c) Exp. D, $\delta\overline{y}$ (K)')
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel('Height (m)')
        ax3.set_title(r'd) Exp. D, $\sigma_{p\prime\prime}/\sigma_{y^b}$ (K$^{-1}$)')
        ax3.set_xlabel('Time (h)')
        ax3.set_ylabel('Height (m)')
        maxtime = exp_D.trun
        ax2.set_xlim([0,maxtime/3600])
        ax3.set_xlim([0,maxtime/3600])
        ax2.set_xticks(np.arange((maxtime+1)/3600))
        ax3.set_xticks(np.arange((maxtime+1)/3600))
        p.colorbar(c0,orientation='horizontal')
        p.colorbar(c1,orientation='horizontal')
        p.colorbar(c2,orientation='horizontal')
        p.colorbar(c3,orientation='horizontal')
        #
        fig.savefig('fig08.png',format='png',dpi=300)
        p.close(fig)

    if fig109:

        path_EAKF = "./runs/seed=181612_enssize=200_EAKF_6hrs/"
        path_LETKF = "./runs/seed=181612_enssize=200_LETKF_6hrs/"

        exp_Aa = pickle.load(open(path_EAKF+"exp_A.pickle", "rb"))
        exp_B1a = pickle.load(open(path_EAKF+"exp_B1.pickle", "rb"))
        exp_B2a = pickle.load(open(path_EAKF+"exp_B2.pickle", "rb"))
        exp_B3a = pickle.load(open(path_EAKF+"exp_B3.pickle", "rb"))
        exp_B4a = pickle.load(open(path_EAKF+"exp_B4.pickle", "rb"))
        exp_C1a = pickle.load(open(path_EAKF+"exp_C1.pickle", "rb"))
        exp_C2a = pickle.load(open(path_EAKF+"exp_C2.pickle", "rb"))
        exp_Da = pickle.load(open(path_EAKF+"exp_D.pickle", "rb"))

        exp_Ab = pickle.load(open(path_LETKF+"exp_A.pickle", "rb"))
        exp_B1b = pickle.load(open(path_LETKF+"exp_B1.pickle", "rb"))
        exp_B2b= pickle.load(open(path_LETKF+"exp_B2.pickle", "rb"))
        exp_B3b = pickle.load(open(path_LETKF+"exp_B3.pickle", "rb"))
        exp_B4b = pickle.load(open(path_LETKF+"exp_B4.pickle", "rb"))
        exp_C1b = pickle.load(open(path_LETKF+"exp_C1.pickle", "rb"))
        exp_C2b = pickle.load(open(path_LETKF+"exp_C2.pickle", "rb"))
        exp_Db = pickle.load(open(path_LETKF+"exp_D.pickle", "rb"))

        fig, [ax1, ax2] = p.subplots(1,2)
        fig.set_size_inches(6,3)
        #
        cc = 0
        linecolors = ['darkviolet','chocolate','darkorange','orange','gold','darkgreen','seagreen','royalblue']
        labels = [r'Exp. A',r'Exp. B$_1$',r'Exp. B$_2$',r'Exp. B$_3$',r'Exp. B$_4$',r'Exp. C$_1$',r'Exp. C$_2$',r'Exp. D']
        for exp in [exp_Aa,exp_B1a,exp_B2a,exp_B3a,exp_B4a,exp_C1a,exp_C2a,exp_Da]:
            # Experiment settings
            true_value = exp.nr.pfac
            parameter_number = exp.nr.parameter_number
            parameter_id = 0
            pt = exp.nr.parameter_transform
            par_tran = exp.da.backgrounds[:,-parameter_number:,:][:,parameter_id,:]
            par_phys = pt[parameter_id](par_tran,kind='inv')
            initpar_phys = pt[parameter_id](exp.da.initial_perturbed_parameters,kind='inv')
            t = exp.da.time/3600.

            # Initial parameter distribution
            ax1.step([0,t[0]],np.mean(initpar_phys,axis=1)*np.array([1,1]),color=linecolors[cc],label=labels[cc])

            # Later times
            ax1.step(t,np.mean(par_phys,axis=1),color=linecolors[cc])
            cc += 1

        cc = 0
        for exp in [exp_Ab,exp_B1b,exp_B2b,exp_B3b,exp_B4b,exp_C1b,exp_C2b,exp_Db]:
            # Experiment settings
            true_value = exp.nr.pfac
            parameter_number = exp.nr.parameter_number
            parameter_id = 0
            pt = exp.nr.parameter_transform
            par_tran = exp.da.backgrounds[:,-parameter_number:,:][:,parameter_id,:]
            par_phys = pt[parameter_id](par_tran,kind='inv')
            initpar_phys = pt[parameter_id](exp.da.initial_perturbed_parameters,kind='inv')
            t = exp.da.time/3600.

            # Initial parameter distribution
            ax2.step([0,t[0]],np.mean(initpar_phys,axis=1)*np.array([1,1]),color=linecolors[cc])

            # Later times
            ax2.step(t,np.mean(par_phys,axis=1),color=linecolors[cc])
            cc += 1

        # Other stuff
        ax1.axhline(y=true_value,linestyle='--',color='black')
        ax1.set_title(r'a) EAKF experiments')
        ax1.set_xlim([0,t.max()])
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel(r'$p$')
        ax1.set_yticks([1,2,3])
        #
        ax2.axhline(y=true_value,linestyle='--',color='black')
        ax2.set_title(r'b) LETKF experiments')
        ax2.set_xlim([0,t.max()])
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel(r'$p$')
        ax2.set_yticks([1,2,3])
        #
        ax2.sharey(ax1)
        fig.legend(loc=8,ncol=4,frameon=False)
        fig.subplots_adjust(bottom=0.35, wspace=0.38)
        #
        fig.savefig('fig09.png',format='png',dpi=300)
        p.close(fig)

    if opt01:

        da_settings = {'cbl_settings' : dict(default_cbl_settings),
                'tspinup' : 10800,
                'trun' : 3600,
                'assimilation_interval' : 3600,
                'obs_coordinates' : np.array([433.]),
                'obs_kinds' : np.array(['theta']),
                'obs_error_sdev_generate' : np.array([.2]),
                'obs_error_sdev_assimilate' : np.array([1.]),
                'localization_cutoff' : 400,
                'nens' : 50,
                'FILTER' : 'EAKF',
                'inflation_rtps_alpha' : 0.1,
                }
        exp = experiment(da_settings)
        plot_CBL_assimilation(exp,'single_assimilation.png')

    if opt02:

        # Make plots
        fig, [ax1, ax2] = p.subplots(1,2,constrained_layout=True)
        fig.set_size_inches(6,3)

        which_cycle = 0
        tt1 = (which_cycle+1)*da_settings_A['assimilation_interval']
        ax1 = plot_CBL_assimilation(exp_A,None,which_cycle=which_cycle,ax=ax1)
        ax1.set_ylim([0,2000])
        ax1.set_xlim([288,298])
        ax1.set_ylabel(r'height (m)')
        ax1.set_title(r'a) Analysis at $t=$%u s'%tt1)
        ax1.legend(frameon=False)

        which_cycle = 31
        tt2 = (which_cycle+1)*da_settings_A['assimilation_interval']
        ax2 = plot_CBL_assimilation(exp_A,None,which_cycle=which_cycle,ax=ax2)
        ax2.set_ylim([0,2000])
        ax2.set_xlim([288,298])
        ax2.set_ylabel(r'height (m)')
        ax2.set_title(r'b) Analysis at $t=$%u s'%tt2)

        fig.savefig('profiles.png',format='png',dpi=300)
        p.close(fig)

    if opt03:
        # Test with "real" observations (no nature run)

        nens = 20
        ############################################################################
        # Control experiment (perfect-model)
        # Create a copy of the default settings
        # and change some of the settings as needed
        cbl_settings_RC = dict(default_cbl_settings)
        cbl_settings_RC['theta_0'] = 290
        cbl_settings_RC['gamma'] = 3e-3
        cbl_settings_RC['Hmax'] = 0.12
        #cbl_settings_RC['parameter_ensemble_min'] = np.array([0.5])
        #cbl_settings_RC['parameter_ensemble_max'] = np.array([2.5])
        da_settings_RC = dict(default_da_settings)
        da_settings_RC['cbl_settings'] = dict(cbl_settings_RC)
        da_settings_RC['nens'] = nens
        da_settings_RC['type'] = 'OSSE'
        da_settings_RC['path'] = "./runs/real_init_exp_seed=%s_enssize=%s_EAKF_6hrs/"%(rnseed,nens)

        # Run and save to disk
        try:
            exp_RC = pickle.load(open(da_settings_RC['path']+"exp_RC.pickle", "rb"))
        except:
            exp_RC = experiment(da_settings_RC)
            setattr(exp_RC,'label','RC')
            pickle.dump(exp_RC, open(da_settings_RC['path']+'exp_RC.pickle', 'wb')) 

        ############################################################################
        # Imperfect-model experiment
        # Create a copy of the default settings
        # and change some of the settings as needed
        cbl_settings_R = dict(default_cbl_settings)
        cbl_settings_R['theta_0'] = 290
        cbl_settings_R['gamma'] = 3e-3
        cbl_settings_R['Hmax'] = 0.12
        #cbl_settings_R['parameter_ensemble_min'] = np.array([0.5])
        #cbl_settings_R['parameter_ensemble_max'] = np.array([2.5])
        da_settings_R = dict(default_da_settings)
        da_settings_R['cbl_settings'] = dict(cbl_settings_R)
        da_settings_R['nens'] = nens
        da_settings_R['type'] = 'real'
        da_settings_R['observation_files'] = './LES_data/Theta/*csv'
        da_settings_R['obs_coordinates'] = None
        da_settings_R['obs_error_sdev_generate'] = None
        da_settings_R['path'] = "./runs/real_init_exp_seed=%s_enssize=%s_EAKF_6hrs/"%(rnseed,nens)

        # Run and save to disk
        try:
            exp_R = pickle.load(open(da_settings_R['path']+'exp_R.pickle', "rb"))
        except:
            exp_R = experiment(da_settings_R)
            setattr(exp_R,'label','R')
            pickle.dump(exp_R, open(da_settings_R['path']+'exp_R.pickle', 'wb')) 

        ############################################################################
        # Run a free ensemble if desired
        freens = False
        if freens:
            maxtime = 3600+21600 # spinup + analysis run
            # Do a free ensemble run (ensemble size set expliclity)
            cbl_settings_free = dict(cbl_settings_R)
            cbl_settings_free['maxtime'] = maxtime
            cbl_settings_free['perturb_ensemble_state'] = False
            cbl_settings_free['parameter_ensemble_min'] = np.array([0.5])
            cbl_settings_free['parameter_ensemble_max'] = np.array([2.5])
            cbl_free = CBL(cbl_settings_free)
            cbl_free.initialize(nens)
            cbl_free.run(output_full_history=True)
            # Make a plot about the free ensemble
            fig, ax1 = p.subplots(1,1,constrained_layout=True)
            fig.set_size_inches(4,3)
            ax1,c1 = plot_spread(cbl_free,plot='mean',ax=ax1)
            ax1.set_ylabel(r'Height (m)')
            ax1.set_title(r'$\overline{\theta}$ (K)')
            ax1.set_xlabel('Time (h)')
            ax1.set_xticks(np.arange((maxtime+1)/3600))
            c1.set_clim([290,296])
            p.colorbar(c1,orientation='horizontal')
            fig.savefig('expR_free_ens.png',format='png',dpi=300)
            p.close(fig)

        #########################################################################################à
        # Make plots: perfect-model experiment (RC)
        fig, [ax1, ax2] = p.subplots(1,2,constrained_layout=True)
        fig.set_size_inches(6,3)

        which_cycle = 0
        tt1 = (which_cycle+1)*da_settings_RC['assimilation_interval']
        ax1 = plot_CBL_assimilation(exp_RC,None,which_cycle=which_cycle,ax=ax1)
        ax1.set_ylim([0,2000])
        ax1.set_xlim([288,298])
        ax1.set_ylabel(r'height (m)')
        ax1.set_title(r'a) Analysis at $t=$%u s'%tt1)
        ax1.legend(frameon=False)

        which_cycle = 31
        tt2 = (which_cycle+1)*da_settings_RC['assimilation_interval']
        ax2 = plot_CBL_assimilation(exp_RC,None,which_cycle=which_cycle,ax=ax2)
        ax2.set_ylim([0,2000])
        ax2.set_xlim([288,298])
        ax2.set_ylabel(r'height (m)')
        ax2.set_title(r'b) Analysis at $t=$%u s'%tt2)

        fig.savefig('expRC_profiles.png',format='png',dpi=300)
        p.close(fig)

        fig, [[ax0, ax1], [ax2, ax3]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(6,6)
        #
        [ax0,ax1,ax2],c0,c1,c2 = plot_CBL_identifiability(exp_RC,da_settings_RC['obs_error_sdev_assimilate'][0],ax=[ax0,ax1,ax2])
        ax0.set_title(r'a) Exp. RC, $\rho(p\prime\prime,y_b}$)')
        ax0.set_xlabel('Time (h)')
        ax0.set_ylabel('Height (m)')
        ax1.set_title(r'b) Exp. RC, $\delta\overline{y}\cdot(\sigma_{p\prime\prime}/\sigma_{y^b})$')
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel('Height (m)')
        ax2.set_title(r'c) Exp. RC, $\delta\overline{p}\prime\prime$')
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel('Height (m)')
        ax3 = plot_CBL_PE(exp_RC,None,ax=ax3)
        ax3.set_title(r'd) Exp. RC, evolution of $p$')
        ax3.set_xlabel('Time (h)')
        ax3.set_yticks([0,1,2,3,4,5])
        maxtime = exp_R.trun
        ax0.set_xlim([0,maxtime/3600])
        ax1.set_xlim([0,maxtime/3600])
        ax2.set_xlim([0,maxtime/3600])
        ax3.set_xlim([0,maxtime/3600])
        ax0.set_xticks(np.arange((maxtime+1)/3600))
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        ax2.set_xticks(np.arange((maxtime+1)/3600))
        ax3.set_xticks(np.arange((maxtime+1)/3600))
        p.colorbar(c0,orientation='horizontal')
        p.colorbar(c1,orientation='horizontal')
        p.colorbar(c2,orientation='horizontal')
        #
        fig.savefig('expRC_PE.png',format='png',dpi=300)
        p.close(fig)
        
        #########################################################################################à
        # Make plots: imperfect-model experiment (R)
        fig, [ax1, ax2] = p.subplots(1,2,constrained_layout=True)
        fig.set_size_inches(6,3)

        which_cycle = 0
        tt1 = (which_cycle+1)*da_settings_R['assimilation_interval']
        ax1 = plot_CBL_assimilation(exp_R,None,which_cycle=which_cycle,ax=ax1)
        ax1.set_ylim([0,2000])
        ax1.set_xlim([288,298])
        ax1.set_ylabel(r'height (m)')
        ax1.set_title(r'a) Analysis at $t=$%u s'%tt1)
        ax1.legend(frameon=False)

        which_cycle = 31
        tt2 = (which_cycle+1)*da_settings_R['assimilation_interval']
        ax2 = plot_CBL_assimilation(exp_R,None,which_cycle=which_cycle,ax=ax2)
        ax2.set_ylim([0,2000])
        ax2.set_xlim([288,298])
        ax2.set_ylabel(r'height (m)')
        ax2.set_title(r'b) Analysis at $t=$%u s'%tt2)

        fig.savefig('expR_profiles.png',format='png',dpi=300)
        p.close(fig)

        fig, [[ax0, ax1], [ax2, ax3]] = p.subplots(2,2,constrained_layout=True)
        fig.set_size_inches(6,6)
        #
        [ax0,ax1,ax2],c0,c1,c2 = plot_CBL_identifiability(exp_R,da_settings_R['obs_error_sdev_assimilate'][0],ax=[ax0,ax1,ax2])
        ax0.set_title(r'a) Exp. R, $\rho(p\prime\prime,y_b}$)')
        ax0.set_xlabel('Time (h)')
        ax0.set_ylabel('Height (m)')
        ax1.set_title(r'b) Exp. R, $\delta\overline{y}\cdot(\sigma_{p\prime\prime}/\sigma_{y^b})$')
        ax1.set_xlabel('Time (h)')
        ax1.set_ylabel('Height (m)')
        ax2.set_title(r'c) Exp. R, $\delta\overline{p}\prime\prime$')
        ax2.set_xlabel('Time (h)')
        ax2.set_ylabel('Height (m)')
        ax3 = plot_CBL_PE(exp_R,None,ax=ax3)
        ax3.set_title(r'd) Exp. R, evolution of $p$')
        ax3.set_xlabel('Time (h)')
        ax3.set_yticks([0,1,2,3,4,5])
        maxtime = exp_R.trun
        ax0.set_xlim([0,maxtime/3600])
        ax1.set_xlim([0,maxtime/3600])
        ax2.set_xlim([0,maxtime/3600])
        ax3.set_xlim([0,maxtime/3600])
        ax0.set_xticks(np.arange((maxtime+1)/3600))
        ax1.set_xticks(np.arange((maxtime+1)/3600))
        ax2.set_xticks(np.arange((maxtime+1)/3600))
        ax3.set_xticks(np.arange((maxtime+1)/3600))
        p.colorbar(c0,orientation='horizontal')
        p.colorbar(c1,orientation='horizontal')
        p.colorbar(c2,orientation='horizontal')
        #
        fig.savefig('expR_PE.png',format='png',dpi=300)
        p.close(fig)

    if opt04:

        # Use only 20 ensemble members for these experiments
        nens = 20

        # Base settings
        cbl_settings_base = dict(default_cbl_settings)
        cbl_settings_base['theta_0'] = 290
        cbl_settings_base['gamma'] = 3e-3
        cbl_settings_base['Hmax'] = 0.12
        da_settings_base = dict(default_da_settings)
        da_settings_base['cbl_settings'] = dict(cbl_settings_base)
        da_settings_base['nens'] = nens
        da_settings_base['type'] = 'real'
        da_settings_base['observation_files'] = './LES_data/Theta/*csv'
        da_settings_base['obs_coordinates'] = None
        da_settings_base['obs_error_sdev_generate'] = None

        # Base experiment (in opt03 has tspinup = 1 hour). Try other lengths.
        da_settings_spinup_1 = dict(da_settings_base)
        da_settings_spinup_1['tspinup'] = 7200
        da_settings_spinup_1['trun'] = 25200-7200
        da_settings_spinup_1['path'] = "./runs/spinup_seed=%s_enssize=%s_EAKF_6hrs/"%(rnseed,nens)

        da_settings_spinup_2 = dict(da_settings_base)
        da_settings_spinup_2['tspinup'] = 10800
        da_settings_spinup_2['trun'] = 25200-10800
        da_settings_spinup_2['path'] = "./runs/spinup_seed=%s_enssize=%s_EAKF_6hrs/"%(rnseed,nens)

        da_settings_spinup_3 = dict(da_settings_base)
        da_settings_spinup_3['tspinup'] = 14400
        da_settings_spinup_3['trun'] = 25200-14400
        da_settings_spinup_3['path'] = "./runs/spinup_seed=%s_enssize=%s_EAKF_6hrs/"%(rnseed,nens)

        # Run and save to disk
        try:
            exp_spinup_1 = pickle.load(open(da_settings_spinup_1['path']+'exp_spinup_1.pickle', "rb"))
        except:
            exp_spinup_1 = experiment(da_settings_spinup_1)
            setattr(exp_spinup_1,'label','spinup_1')
            pickle.dump(exp_spinup_1, open(da_settings_spinup_1['path']+'exp_spinup_1.pickle', 'wb')) 
        try:
            exp_spinup_2 = pickle.load(open(da_settings_spinup_2['path']+'exp_spinup_2.pickle', "rb"))
        except:
            exp_spinup_2 = experiment(da_settings_spinup_2)
            setattr(exp_spinup_2,'label','spinup_2')
            pickle.dump(exp_spinup_2, open(da_settings_spinup_2['path']+'exp_spinup_2.pickle', 'wb')) 
        try:
            exp_spinup_3 = pickle.load(open(da_settings_spinup_3['path']+'exp_spinup_3.pickle', "rb"))
        except:
            exp_spinup_3 = experiment(da_settings_spinup_3)
            setattr(exp_spinup_3,'label','spinup_3')
            pickle.dump(exp_spinup_3, open(da_settings_spinup_3['path']+'exp_spinup_3.pickle', 'wb')) 

        for exp,settings in zip([exp_spinup_1,exp_spinup_2,exp_spinup_3],[da_settings_spinup_1,da_settings_spinup_2,da_settings_spinup_3]):
            label = exp.label
            maxtime = exp.trun

            fig, [[ax0, ax1], [ax2, ax3]] = p.subplots(2,2,constrained_layout=True)
            fig.set_size_inches(6,6)
            #
            [ax0,ax1,ax2],c0,c1,c2 = plot_CBL_identifiability(exp,settings['obs_error_sdev_assimilate'][0],ax=[ax0,ax1,ax2])
            ax0.set_title(r'a) Exp. %s, $\rho(p\prime\prime,y_b}$)'%label)
            ax0.set_xlabel('Time (h)')
            ax0.set_ylabel('Height (m)')
            ax1.set_title(r'b) Exp. %s, $\delta\overline{y}\cdot(\sigma_{p\prime\prime}/\sigma_{y^b})$'%label)
            ax1.set_xlabel('Time (h)')
            ax1.set_ylabel('Height (m)')
            ax2.set_title(r'c) Exp. %s, $\delta\overline{p}\prime\prime$'%label)
            ax2.set_xlabel('Time (h)')
            ax2.set_ylabel('Height (m)')
            ax3 = plot_CBL_PE(exp,None,ax=ax3)
            ax3.set_title(r'd) Exp. %s, evolution of $p$'%label)
            ax3.set_xlabel('Time (h)')
            ax3.set_yticks([0,1,2,3,4,5])
            ax0.set_xlim([0,maxtime/3600])
            ax1.set_xlim([0,maxtime/3600])
            ax2.set_xlim([0,maxtime/3600])
            ax3.set_xlim([0,maxtime/3600])
            ax0.set_xticks(np.arange((maxtime+1)/3600))
            ax1.set_xticks(np.arange((maxtime+1)/3600))
            ax2.set_xticks(np.arange((maxtime+1)/3600))
            ax3.set_xticks(np.arange((maxtime+1)/3600))
            p.colorbar(c0,orientation='horizontal')
            p.colorbar(c1,orientation='horizontal')
            p.colorbar(c2,orientation='horizontal')
            #
            fig.savefig('exp_%s_PE.png'%label,format='png',dpi=300)
            p.close(fig)

    check_reproducibility = False
    if check_reproducibility:
        path = default_da_settings['path']
        for i in ["exp_A.pickle",
                  "exp_A_noPE.pickle",
                  "exp_B1.pickle",
                  "exp_B2.pickle",
                  "exp_B3.pickle",
                  "exp_B4.pickle",
                  "exp_C1.pickle",
                  "exp_C2.pickle",
                  "exp_D.pickle"]:
            experiment = pickle.load(open(path+i, "rb"))
            print(i)
            print("OBS       :",experiment.da.observations[:,0])
            print("PARAMETERS:",experiment.da.initial_perturbed_parameters[0,:5])
            print("STATE     :",experiment.da.initial_state[0,0,:5])

